# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum

class NtStatus(enum.IntEnum):
    """Known NtStatus return values."""

    SEC_I_COMPLETE_AND_CONTINUE = ...
    """
    The client must call :meth:`complete_auth_token`, pass the output to the
    sever and wait for a return token to process.
    """
    SEC_I_COMPLETE_NEEDED = ...
    """
    The client must finish building the message and then call
    :meth:`complete_auth_token`.
    """
    SEC_I_CONTINUE_NEEDED = ...
    """
    The client must send the output token to the server and wait for a return
    token to process.
    """
    SEC_E_OK = ...
    """
    The security context was successfully initialized.
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import socket
import typing as t

import pytest

import sspi


@pytest.fixture()
def authenticated_contexts() -> tuple[sspi.InitiatorSecurityContext, sspi.AcceptorSecurityContext]:
    spn = f"host/{socket.gethostname()}"
    c_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_OUTBOUND)
    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    isc_req = (
        sspi.IscReq.ISC_REQ_ALLOCATE_MEMORY
        | sspi.IscReq.ISC_REQ_CONFIDENTIALITY
        | sspi.IscReq.ISC_REQ_INTEGRITY
        | sspi.IscReq.ISC_REQ_SEQUENCE_DETECT
        | sspi.IscReq.ISC_REQ_REPLAY_DETECT
    )
    asc_req = (
        sspi.AscReq.ASC_REQ_ALLOCATE_MEMORY
        | sspi.AscReq.ASC_REQ_CONFIDENTIALITY
        | sspi.AscReq.ASC_REQ_INTEGRITY
        | sspi.AscReq.ASC_REQ_SEQUENCE_DETECT
        | sspi.AscReq.ASC_REQ_REPLAY_DETECT
    )

    c_ctx = s_ctx = server_token = None
    while True:
        c_input_buffers = None
        if server_token:
            c_input_buffers = sspi.SecBufferDesc(
                [
                    sspi.SecBuffer(server_token, sspi.SecBufferType.SECBUFFER_TOKEN),
                ]
            )

        c_output_buffers = sspi.SecBufferDesc(
            [
                sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
            ]
        )
        c_ctx, c_status = sspi.initialize_security_context(
            credential=c_cred,
            context=c_ctx,
            target_name=spn,
            context_req=isc_req,
            target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
            input_buffers=c_input_buffers,
            output_buffers=c_output_buffers,
        )

        if c_output_buffers[0].count:
            s_input_buffers = sspi.SecBufferDesc(
                [
                    sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
                ]
            )
            s_output_buffers = sspi.SecBufferDesc(
                [
                    sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
                ]
            )
            s_ctx, s_status = sspi.accept_security_context(
                credential=s_cred,
                context=s_ctx,
                input_buffers=s_input_buffers,
                context_req=asc_req,
                target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
                output_buffers=s_output_buffers,
            )

            if s_output_buffers[0].count:
                server_token = bytearray(s_output_buffers[0].data)
            elif s_status == sspi.NtStatus.SEC_E_OK:
                break
            else:
                raise ValueError(f"Expected SEC_E_OK but got {c_status.name}")

        elif c_status == sspi.NtStatus.SEC_E_OK:
            break
        else:
            raise ValueError(f"Expected SEC_E_OK but got {c_status.name}")

    return c_ctx, t.cast(sspi.AcceptorSecurityContext, s_ctx)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum

from cpython.exc cimport PyErr_SetFromWindowsErr
from cpython.unicode cimport PyUnicode_GET_LENGTH
from libc.stdlib cimport free, malloc

from sspi._credential cimport Credential
from sspi._text cimport WideCharString, wide_char_to_str
from sspi._win32_types cimport *


cdef extern from "Security.h":
    unsigned long SECPKG_CRED_ATTR_NAMES
    unsigned long SECPKG_CRED_ATTR_SSI_PROVIDER
    unsigned long SECPKG_CRED_ATTR_KDC_PROXY_SETTINGS
    unsigned long SECPKG_CRED_ATTR_CERT
    unsigned long SECPKG_CRED_ATTR_PAC_BYPASS

    unsigned long KDC_PROXY_SETTINGS_V1
    unsigned long _KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY "KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY"

    cdef struct _SecPkgCredentials_KdcProxySettingsW:
        ULONG Version
        ULONG Flags
        USHORT ProxyServerOffset
        USHORT ProxyServerLength
        USHORT ClientTlsCredOffset
        USHORT ClientTlsCredLength
    ctypedef _SecPkgCredentials_KdcProxySettingsW SecPkgCredentials_KdcProxySettingsW
    ctypedef SecPkgCredentials_KdcProxySettingsW *PSecPkgCredentials_KdcProxySettingsW

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-setcredentialsattributesw
    SECURITY_STATUS SetCredentialsAttributesW(
        PCredHandle   phCredential,
        unsigned long ulAttribute,
        void          *pBuffer,
        unsigned long cbBuffer
    ) nogil


class KdcProxySettingsFlags(enum.IntEnum):
    KDX_PROXY_SETTINGS_FLAGS_NONE = 0
    KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY = _KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY

cdef class SecPkgCred:

    cdef (unsigned long, void *, unsigned long) __c_value__(SecPkgCred self):
        return (0, NULL, 0)

cdef class SecPkgCredKdcProxySettings(SecPkgCred):
    cdef PSecPkgCredentials_KdcProxySettingsW raw

    def __cinit__(
        SecPkgCredKdcProxySettings self,
        *,
        flags: KdcProxySettingsFlags | int = 0,
        proxy_server: str | None = None,
    ):
        cdef WideCharString proxy_wchar = WideCharString(proxy_server)
        cdef int proxy_wchar_len = proxy_wchar.length * sizeof(wchar_t)
        cdef unsigned char[:] temp_ptr = None

        raw_length = sizeof(SecPkgCredentials_KdcProxySettingsW) + proxy_wchar_len
        self.raw = <PSecPkgCredentials_KdcProxySettingsW>malloc(raw_length)
        if not self.raw:
            raise MemoryError("Cannot malloc SecPkgCredKdcProxySettings buffers")

        cdef unsigned char[:] raw_ptr = <unsigned char[:raw_length]><unsigned char*>self.raw
        offset = sizeof(SecPkgCredentials_KdcProxySettingsW)

        self.raw.Version = KDC_PROXY_SETTINGS_V1
        self.raw.Flags = flags

        self.raw.ProxyServerOffset = 0
        self.raw.ProxyServerLength = proxy_wchar_len
        if proxy_wchar_len:
            self.raw.ProxyServerOffset = offset

            temp_ptr = <unsigned char[:proxy_wchar_len]><unsigned char*>proxy_wchar.raw
            raw_ptr[offset:offset + proxy_wchar_len] = temp_ptr.copy()
            offset += proxy_wchar_len

        self.raw.ClientTlsCredOffset = 0
        self.raw.ClientTlsCredLength = 0

    def __dealloc__(SecPkgCredKdcProxySettings self):
        if self.raw:
            free(self.raw)
            self.raw = NULL

    cdef (unsigned long, void *, unsigned long) __c_value__(SecPkgCredKdcProxySettings self):
        cdef unsigned long length = sizeof(SecPkgCredentials_KdcProxySettingsW) + \
            self.raw.ProxyServerLength + \
            self.raw.ClientTlsCredLength
        return (SECPKG_CRED_ATTR_KDC_PROXY_SETTINGS, self.raw, length)

    def __repr__(SecPkgCredKdcProxySettings self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'flags': self.flags,
            'proxy_server': repr(self.proxy_server),
        }.items()]

        return f"SecPkgCredKdcProxySettings({', '.join(kwargs)})"

    @property
    def version(SecPkgCredKdcProxySettings self) -> int:
        return self.raw.Version

    @property
    def flags(SecPkgCredKdcProxySettings self) -> KdcProxySettingsFlags:
        return KdcProxySettingsFlags(self.raw.Flags)

    @property
    def proxy_server(SecPkgCredKdcProxySettings self) -> str | None:
        cdef const wchar_t *raw = NULL
        cdef int size = -1

        cdef unsigned char [:] raw_view = None
        if self.raw.ProxyServerOffset and self.raw.ProxyServerLength:
            raw_view = <unsigned char [:self.raw.ProxyServerOffset + self.raw.ProxyServerLength]> \
                <unsigned char*>self.raw

            raw = <wchar_t*>&raw_view[self.raw.ProxyServerOffset]
            size = self.raw.ProxyServerLength // sizeof(wchar_t)

        return wide_char_to_str(raw, size)

def set_credentials_attributes(
    Credential credential not None,
    SecPkgCred attribute not None,
) -> None:
    cdef (unsigned long, void*, unsigned long) raw = attribute.__c_value__()

    with nogil:
        res = SetCredentialsAttributesW(
            &credential.raw,
            raw[0],
            raw[1],
            raw[2],
        )

    if res:
        PyErr_SetFromWindowsErr(res)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import pytest

import sspi


def test_get_outbound_with_default_context() -> None:
    cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
    )
    assert isinstance(cred, sspi.Credential)
    assert cred.expiry > 0


def test_get_inbound_with_default_context() -> None:
    cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_INBOUND,
    )
    assert isinstance(cred, sspi.Credential)
    assert cred.expiry > 0


@pytest.mark.parametrize(
    ["username", "domain", "expected_str", "expected_repr"],
    [
        (
            "username",
            None,
            "username",
            "WinNTAuthIdentity(username='username', domain=None, password='pass', flags=2, package_list=None)",
        ),
        (
            "username@DOMAIN.COM",
            None,
            "username@DOMAIN.COM",
            "WinNTAuthIdentity(username='username@DOMAIN.COM', domain=None, password='pass', flags=2, package_list=None)",
        ),
        (
            "username",
            "DOMAIN",
            "DOMAIN\\username",
            "WinNTAuthIdentity(username='username', domain='DOMAIN', password='pass', flags=2, package_list=None)",
        ),
        (
            "user\U0001F4A9",
            "DOMAIN\U0001F4A9",
            "DOMAIN\U0001F4A9\\user\U0001F4A9",
            "WinNTAuthIdentity(username='user\U0001F4A9', domain='DOMAIN\U0001F4A9', password='pass', flags=2, package_list=None)",
        ),
    ],
    ids=[
        "username_only",
        "username_domain_upn",
        "username_domain_netbios",
        "unicode_surrogates",
    ],
)
def test_get_cred_with_explicit_credentials(
    username: str | None,
    domain: str | None,
    expected_str: str,
    expected_repr: str,
) -> None:
    password = "pass"
    auth_identity = sspi.WinNTAuthIdentity(
        username=username,
        domain=domain,
        password=password,
    )
    assert auth_identity.username == username
    assert auth_identity.domain == domain
    assert auth_identity.password == password
    assert auth_identity.package_list is None
    assert auth_identity.flags == sspi.WinNTAuthFlags.SEC_WINNT_AUTH_IDENTITY_UNICODE
    assert str(auth_identity) == expected_str
    assert repr(auth_identity) == expected_repr

    cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_identity,
    )
    assert isinstance(cred, sspi.Credential)
    assert cred.expiry > 0


def test_get_cred_with_explicit_credentials_package_list() -> None:
    auth_identity = sspi.WinNTAuthIdentity(
        username="username",
        password="DummyPasswordHere",
        package_list="!kerberos",
    )
    assert auth_identity.package_list == "!kerberos"
    assert auth_identity.flags == sspi.WinNTAuthFlags.SEC_WINNT_AUTH_IDENTITY_UNICODE

    cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_identity,
    )
    assert isinstance(cred, sspi.Credential)
    assert cred.expiry > 0


def test_get_credential_with_principal() -> None:
    cred = sspi.acquire_credentials_handle(
        "Principal",
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
    )
    assert isinstance(cred, sspi.Credential)
    assert cred.expiry > 0


def test_fail_with_invalid_package_name() -> None:
    with pytest.raises(WindowsError) as e:
        sspi.acquire_credentials_handle(None, "Invalid", sspi.CredentialUse.SECPKG_CRED_OUTBOUND)

    assert e.value.winerror == -2146893051  # SEC_E_SECPKG_NOT_FOUND



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import sspi


def test_query_names(
    authenticated_contexts: tuple[sspi.InitiatorSecurityContext, sspi.AcceptorSecurityContext],
) -> None:
    c_actual = sspi.query_context_attributes(authenticated_contexts[0], sspi.SecPkgContextNames)
    assert isinstance(c_actual, sspi.SecPkgContextNames)
    assert isinstance(c_actual.username, str)

    s_actual = sspi.query_context_attributes(authenticated_contexts[1], sspi.SecPkgContextNames)
    assert isinstance(s_actual, sspi.SecPkgContextNames)
    assert isinstance(s_actual.username, str)

    assert c_actual.username == s_actual.username
    assert repr(c_actual) == repr(s_actual)


def test_query_package_info(
    authenticated_contexts: tuple[sspi.InitiatorSecurityContext, sspi.AcceptorSecurityContext],
) -> None:
    c_actual = sspi.query_context_attributes(authenticated_contexts[0], sspi.SecPkgContextPackageInfo)
    assert isinstance(c_actual, sspi.SecPkgContextPackageInfo)
    assert isinstance(c_actual.capabilities, sspi.SecurityPackageCapability)
    assert isinstance(c_actual.version, int)
    assert isinstance(c_actual.rpcid, int)
    assert isinstance(c_actual.max_token, int)
    assert isinstance(c_actual.name, str)
    assert isinstance(c_actual.comment, str)

    s_actual = sspi.query_context_attributes(authenticated_contexts[1], sspi.SecPkgContextPackageInfo)
    assert isinstance(s_actual, sspi.SecPkgContextPackageInfo)
    assert isinstance(s_actual.capabilities, sspi.SecurityPackageCapability)
    assert isinstance(s_actual.version, int)
    assert isinstance(s_actual.rpcid, int)
    assert isinstance(s_actual.max_token, int)
    assert isinstance(s_actual.name, str)
    assert isinstance(s_actual.comment, str)

    assert c_actual.capabilities == s_actual.capabilities
    assert c_actual.version == s_actual.version
    assert c_actual.rpcid == s_actual.rpcid
    assert c_actual.max_token == s_actual.max_token
    assert c_actual.name == s_actual.name
    assert c_actual.comment == s_actual.comment
    assert repr(c_actual) == repr(s_actual)


def test_query_session_keys(
    authenticated_contexts: tuple[sspi.InitiatorSecurityContext, sspi.AcceptorSecurityContext],
) -> None:
    c_actual = sspi.query_context_attributes(authenticated_contexts[0], sspi.SecPkgContextSessionKey)
    assert isinstance(c_actual, sspi.SecPkgContextSessionKey)
    assert isinstance(c_actual.session_key, bytes)

    s_actual = sspi.query_context_attributes(authenticated_contexts[1], sspi.SecPkgContextSessionKey)
    assert isinstance(s_actual, sspi.SecPkgContextSessionKey)
    assert isinstance(s_actual.session_key, bytes)

    assert c_actual.session_key == s_actual.session_key
    assert repr(c_actual) == repr(s_actual)


def test_query_sizes(
    authenticated_contexts: tuple[sspi.InitiatorSecurityContext, sspi.AcceptorSecurityContext],
) -> None:
    c_actual = sspi.query_context_attributes(authenticated_contexts[0], sspi.SecPkgContextSizes)
    assert isinstance(c_actual, sspi.SecPkgContextSizes)
    assert isinstance(c_actual.max_signature, int)
    assert isinstance(c_actual.max_token, int)
    assert isinstance(c_actual.block_size, int)
    assert isinstance(c_actual.security_trailer, int)

    s_actual = sspi.query_context_attributes(authenticated_contexts[1], sspi.SecPkgContextSizes)
    assert isinstance(s_actual, sspi.SecPkgContextSizes)
    assert isinstance(s_actual.max_signature, int)
    assert isinstance(s_actual.max_token, int)
    assert isinstance(s_actual.block_size, int)
    assert isinstance(s_actual.security_trailer, int)

    assert c_actual.max_signature == s_actual.max_signature
    assert c_actual.max_token == s_actual.max_token
    assert c_actual.block_size == s_actual.block_size
    assert c_actual.security_trailer == s_actual.security_trailer
    assert repr(c_actual) == repr(s_actual)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import collections
import enum

from cpython.exc cimport PyErr_SetFromWindowsErr

from sspi._security_buffer cimport FreeContextBuffer
from sspi._security_context cimport SecurityContext
from sspi._security_package cimport PSecPkgInfoW
from sspi._text cimport wide_char_to_str
from sspi._win32_types cimport *

from sspi._ntstatus import NtStatus
from sspi._security_package import SecurityPackageCapability


cdef extern from "Security.h":
    unsigned long SECPKG_ATTR_SIZES
    unsigned long SECPKG_ATTR_NAMES
    unsigned long SECPKG_ATTR_LIFESPAN
    unsigned long SECPKG_ATTR_DCE_INFO
    unsigned long SECPKG_ATTR_STREAM_SIZES
    unsigned long SECPKG_ATTR_KEY_INFO
    unsigned long SECPKG_ATTR_AUTHORITY
    unsigned long SECPKG_ATTR_PROTO_INFO
    unsigned long SECPKG_ATTR_PASSWORD_EXPIRY
    unsigned long SECPKG_ATTR_SESSION_KEY
    unsigned long SECPKG_ATTR_PACKAGE_INFO
    unsigned long SECPKG_ATTR_USER_FLAGS
    unsigned long SECPKG_ATTR_NEGOTIATION_INFO
    unsigned long SECPKG_ATTR_NATIVE_NAMES
    unsigned long SECPKG_ATTR_FLAGS
    unsigned long SECPKG_ATTR_USE_VALIDATED
    unsigned long SECPKG_ATTR_CREDENTIAL_NAME
    unsigned long SECPKG_ATTR_TARGET_INFORMATION
    unsigned long SECPKG_ATTR_ACCESS_TOKEN
    unsigned long SECPKG_ATTR_TARGET
    unsigned long SECPKG_ATTR_AUTHENTICATION_ID
    unsigned long SECPKG_ATTR_LOGOFF_TIME
    unsigned long SECPKG_ATTR_NEGO_KEYS
    unsigned long SECPKG_ATTR_PROMPTING_NEEDED
    unsigned long SECPKG_ATTR_UNIQUE_BINDINGS
    unsigned long SECPKG_ATTR_ENDPOINT_BINDINGS
    unsigned long SECPKG_ATTR_CLIENT_SPECIFIED_TARGET
    unsigned long SECPKG_ATTR_LAST_CLIENT_TOKEN_STATUS
    unsigned long SECPKG_ATTR_NEGO_PKG_INFO
    unsigned long SECPKG_ATTR_NEGO_STATUS
    unsigned long SECPKG_ATTR_CONTEXT_DELETED
    unsigned long SECPKG_ATTR_DTLS_MTU
    unsigned long SECPKG_ATTR_DATAGRAM_SIZES
    unsigned long SECPKG_ATTR_SUBJECT_SECURITY_ATTRIBUTES
    unsigned long SECPKG_ATTR_APPLICATION_PROTOCOL
    unsigned long SECPKG_ATTR_NEGOTIATED_TLS_EXTENSIONS
    unsigned long SECPKG_ATTR_IS_LOOPBACK

    cdef struct _SecPkgContext_NamesW:
        LPWSTR sUserName
    ctypedef _SecPkgContext_NamesW SecPkgContext_NamesW
    ctypedef SecPkgContext_NamesW *PSecPkgContext_NamesW

    cdef struct _SecPkgContext_PackageInfoW:
        PSecPkgInfoW PackageInfo
    ctypedef _SecPkgContext_PackageInfoW SecPkgContext_PackageInfoW
    ctypedef SecPkgContext_PackageInfoW *PSecPkgContext_PackageInfoW

    cdef struct _SecPkgContext_Sizes:
        unsigned long cbMaxToken
        unsigned long cbMaxSignature
        unsigned long cbBlockSize
        unsigned long cbSecurityTrailer
    ctypedef _SecPkgContext_Sizes SecPkgContext_Sizes
    ctypedef SecPkgContext_Sizes *PSecPkgContext_Sizes

    cdef struct _SecPkgContext_SessionKey:
        unsigned long SessionKeyLength
        unsigned char *SessionKey
    ctypedef _SecPkgContext_SessionKey SecPkgContext_SessionKey
    ctypedef SecPkgContext_SessionKey *PSecPkgContext_SessionKey

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-querycontextattributesw
    SECURITY_STATUS QueryContextAttributesW(
        PCtxtHandle   phContext,
        unsigned long ulAttribute,
        void          *pBuffer
    ) nogil

cdef class SecPkgContext:

    cdef (unsigned long, void *) __c_value__(SecPkgContext self):
        return (0, NULL)

cdef class SecPkgContextNames(SecPkgContext):
    cdef SecPkgContext_NamesW raw

    def __dealloc__(SecPkgContextNames self):
        if self.raw.sUserName:
            FreeContextBuffer(self.raw.sUserName)
            self.raw.sUserName = NULL

    cdef (unsigned long, void *) __c_value__(SecPkgContextNames self):
        return (SECPKG_ATTR_NAMES, &self.raw)

    def __repr__(SecPkgContextNames self):
        return f"SecPkgContextNames(username={self.username!r})"

    @property
    def username(SecPkgContextNames self) -> str:
        if self.raw.sUserName == NULL:
            return ""
        else:
            return wide_char_to_str(self.raw.sUserName)

cdef class SecPkgContextPackageInfo(SecPkgContext):
    cdef SecPkgContext_PackageInfoW raw

    def __dealloc__(SecPkgContextPackageInfo self):
        if self.raw.PackageInfo:
            FreeContextBuffer(self.raw.PackageInfo)
            self.raw.PackageInfo = NULL

    cdef (unsigned long, void *) __c_value__(SecPkgContextPackageInfo self):
        return (SECPKG_ATTR_PACKAGE_INFO, &self.raw)

    def __repr__(SecPkgContextPackageInfo self):
        kwargs = [f"{k}={v}" for k, v in {
            'capabilities': self.capabilities.value,
            'version': self.version,
            'rpcid': self.rpcid,
            'max_token': self.max_token,
            'name': repr(self.name),
            'comment': repr(self.comment),
        }.items()]

        return f"SecPkgContextPackageInfo({', '.join(kwargs)})"

    @property
    def capabilities(SecPkgContextPackageInfo self) -> SecurityPackageCapability:
        return SecurityPackageCapability(self.raw.PackageInfo.fCapabilities)

    @property
    def version(SecPkgContextPackageInfo self) -> int:
        return self.raw.PackageInfo.wVersion

    @property
    def rpcid(SecPkgContextPackageInfo self) -> int:
        return self.raw.PackageInfo.wRPCID

    @property
    def max_token(SecPkgContextPackageInfo self) -> int:
        return self.raw.PackageInfo.cbMaxToken

    @property
    def name(SecPkgContextPackageInfo self) -> str:
        return wide_char_to_str(self.raw.PackageInfo.Name, size=-1, none_is_empty=1)

    @property
    def comment(SecPkgContextPackageInfo self) -> str:
        return wide_char_to_str(self.raw.PackageInfo.Comment, size=-1, none_is_empty=1)

cdef class SecPkgContextSessionKey(SecPkgContext):
    cdef SecPkgContext_SessionKey raw

    def __dealloc__(SecPkgContextSessionKey self):
        if self.raw.SessionKey:
            FreeContextBuffer(self.raw.SessionKey)
            self.raw.SessionKeyLength = 0
            self.raw.SessionKey = NULL

    cdef (unsigned long, void *) __c_value__(SecPkgContextSessionKey self):
        return (SECPKG_ATTR_SESSION_KEY, &self.raw)

    def __repr__(SecPkgContextSessionKey self):
        return f"SecPkgContextSessionKey(session_key={self.session_key!r})"

    @property
    def session_key(SecPkgContextSessionKey self) -> bytes:
        if self.raw.SessionKeyLength and self.raw.SessionKey != NULL:
            return (<char *>self.raw.SessionKey)[:self.raw.SessionKeyLength]
        else:
            return b""

cdef class SecPkgContextSizes(SecPkgContext):
    cdef SecPkgContext_Sizes raw

    cdef (unsigned long, void *) __c_value__(SecPkgContextSizes self):
        return (SECPKG_ATTR_SIZES, &self.raw)

    def __repr__(SecPkgContextSizes self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'max_token': self.max_token,
            'max_signature': self.max_signature,
            'block_size': self.block_size,
            'security_trailer': self.security_trailer,
        }.items()]

        return f"SecPkgContextSizes({', '.join(kwargs)})"

    @property
    def max_token(SecPkgContextSizes self) -> int:
        return self.raw.cbMaxToken

    @property
    def max_signature(SecPkgContextSizes self) -> int:
        return self.raw.cbMaxSignature

    @property
    def block_size(SecPkgContextSizes self) -> int:
        return self.raw.cbBlockSize

    @property
    def security_trailer(SecPkgContextSizes self) -> int:
        return self.raw.cbSecurityTrailer

def query_context_attributes(
    SecurityContext context not None,
    type attribute not None,
) -> SecPkgContext:
    if not issubclass(attribute, SecPkgContext):
        raise TypeError("attribute must be a type of SecPkgContext")

    cdef SecPkgContext value = attribute()
    cdef (unsigned long, void*) raw = value.__c_value__()

    with nogil:
        res = QueryContextAttributesW(
            &context.raw,
            raw[0],
            raw[1],
        )

    if res:
        PyErr_SetFromWindowsErr(res)

    return value



# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stddef cimport size_t, wchar_t


cdef extern from "Windows.h":
    ctypedef void *PVOID;

    ctypedef unsigned short USHORT

    ctypedef long LONG
    ctypedef unsigned long ULONG
    ctypedef LONG SECURITY_STATUS

    ctypedef wchar_t WCHAR
    ctypedef WCHAR *LPWSTR
    ctypedef WCHAR SEC_WCHAR;

    ctypedef size_t ULONG_PTR

    cdef struct _SecHandle:
        ULONG_PTR dwLower
        ULONG_PTR dwUpper
    ctypedef _SecHandle SecHandle
    ctypedef SecHandle *PSecHandle

    ctypedef SecHandle CredHandle
    ctypedef PSecHandle PCredHandle

    ctypedef SecHandle CtxtHandle
    ctypedef PSecHandle PCtxtHandle

    cdef struct _SECURITY_INTEGER:
        unsigned long LowPart
        long          HighPart
    ctypedef _SECURITY_INTEGER SECURITY_INTEGER
    ctypedef SECURITY_INTEGER *PSECURITY_INTEGER
    ctypedef SECURITY_INTEGER TimeStamp
    ctypedef SECURITY_INTEGER *PTimeStamp



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import dataclasses
import enum

class SecurityPackageCapability(enum.IntFlag):
    """Security Package capability flags."""

    SECPKG_FLAG_INTEGRITY = ...
    """Supports the :meth:`make_signature` and :meth:`verify_signature` functions."""
    SECPKG_FLAG_PRIVACY = ...
    """Supports the :meth:`encrypt_message` and :meth:`decrypt_message` functions."""
    SECPKG_FLAG_TOKEN_ONLY = ...
    """Package is only interested in security-token porition of messages."""
    SECPKG_FLAG_DATAGRAM = ...
    """Supports datagram-style authentication."""
    SECPKG_FLAG_CONNECTION = ...
    """Supports connection-oriented style authentication."""
    SECPKG_FLAG_MULTI_REQUIRED = ...
    """Multiple legs are required for authentication."""
    SECPKG_FLAG_CLIENT_ONLY = ...
    """Server authentication support is not provided."""
    SECPKG_FLAG_EXTENDED_ERROR = ...
    """Supported extended error handling."""
    SECPKG_FLAG_IMPERSONATION = ...
    """Supports Windows impersonation in server contexts."""
    SECPKG_FLAG_ACCEPT_WIN32_NAME = ...
    """Understands Windows principal and target names."""
    SECPKG_FLAG_STREAM = ...
    """Supports stream semantics."""
    SECPKG_FLAG_NEGOTIABLE = ...
    """Can be used by the Negotiate security package."""
    SECPKG_FLAG_GSS_COMPATIBLE = ...
    """Supports GSS compatibility."""
    SECPKG_FLAG_LOGON = ...
    """Supports LsaLogonUser."""
    SECPKG_FLAG_ASCII_BUFFERS = ...
    """Token buffers are in ASCII character format."""
    SECPKG_FLAG_FRAGMENT = ...
    """Supports separating large tokens into smaller buffers."""
    SECPKG_FLAG_MUTUAL_AUTH = ...
    """Supports mutual authentication."""
    SECPKG_FLAG_DELEGATION = ...
    """Supports delegation."""
    SECPKG_FLAG_READONLY_WITH_CHECKSUM = ...
    """Supports using a checksum instead of in-place encryption with :meth:`encrypt_message`."""
    SECPKG_FLAG_RESTRICTED_TOKENS = ...
    """Supports callers with restricted tokens."""
    SECPKG_FLAG_NEGO_EXTENDER = ...
    """Extends the Negotiate security package."""
    SECPKG_FLAG_NEGOTIABLE2 = ...
    """Is negotiated by the nego extended package."""
    SECPKG_FLAG_APPCONTAINER_PASSTHROUGH = ...
    """Receieves all calls from app container apps."""
    SECPKG_FLAG_APPCONTAINER_CHECKS = ...
    """Receives calls from app containers apps for more specific scenarios."""
    SECPKG_FLAG_CREDENTIAL_ISOLATION_ENABLED = ...
    """Is running with Credential Guard enabled."""
    SECPKG_FLAG_APPLY_LOOPBACK = ...
    """Supports reliable detection of loopback."""

@dataclasses.dataclass
class SecPkgInfo:
    """Security Package info returned by :meth:`enumerate_security_packages`."""

    capabilities: SecurityPackageCapability
    """Set of capabilities of the security package."""
    version: int
    """The version of the package."""
    rpcid: int
    """The DCE RPC identifier if appropriate."""
    max_token: int
    """The maximum size, in bytes, of the token."""
    name: str
    """The name of the security package."""
    comment: str
    """Additional information about the security package."""

def enumerate_security_packages() -> list[SecPkgInfo]:
    """Enumerates installed security packages.

    Enumerates the security packages that are available to the client.

    The ``name`` attribute is the security package name that can be used with
    :meth:`acquire_credentials_handle`.

    This wraps the `EnumerateSecurityPackages`_ Win32 function.

    Returns:
        list[SecPkgInfo]: The list of security packages.

    Raises:
        WindowsError: If the function failed.

    .. EnumerateSecurityPackages:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-enumeratesecuritypackagesw
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import typing as t

from ._security_context import SecurityContext
from ._security_package import SecurityPackageCapability

T = t.TypeVar("T", bound=SecPkgContext)

class SecPkgContext:
    """Base class for context attribute types."""

class SecPkgContextNames(SecPkgContext):
    """Security Package Names

    The structure indicates the name of the user associated with a security
    context.

    This wraps the `SecPkgContext_NamesW`_ Win32 structure.

    .. _SecPkgContext_NamesW:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-secpkgcontext_namesw
    """

    @property
    def username(self) -> str:
        """The user represented by the context."""

class SecPkgContextPackageInfo(SecPkgContext):
    """Security Package Information

    The structure contains the package information associated with a security
    context.

    This wraps the `SecPkgContext_PackageInfoW`_ Win32 structure.

    .. _SecPkgContext_PackageInfoW:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-secpkgcontext_packageinfow
    """

    @property
    def capabilities(self) -> SecurityPackageCapability:
        """Set of capabilities of the security package."""
    @property
    def version(self) -> int:
        """The version of the package."""
    @property
    def rpcid(self) -> int:
        """The DCE RPC identifier if appropriate."""
    @property
    def max_token(self) -> int:
        """The maximum size, in bytes, of the token."""
    @property
    def name(self) -> str:
        """The name of the security package."""
    @property
    def comment(self) -> str:
        """Additional information about the security package."""

class SecPkgContextSessionKey(SecPkgContext):
    """Security Package session key.

    The structure contains information about the session key used for the
    security context.

    This wraps the `SecPkgContext_SessionKey`_ Win32 structure.

    .. _SecPkgContext_SessionKey:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-secpkgcontext_sessionkey
    """

    @property
    def session_key(self) -> bytes:
        """The session key for the security context."""

class SecPkgContextSizes(SecPkgContext):
    """Security Package sizes.

    The structure indicates the sizes of important structures used in the
    message support functions. Use :meth:`query_context_attributes` to generate
    this instance.

    This wraps the `SecPkgContext_Sizes`_ Win32 structure.

    .. _SecPkgContextSizes:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-secpkgcontext_sizes
    """

    @property
    def max_token(self) -> int:
        """Maximum security token sizes."""
    @property
    def max_signature(self) -> int:
        """Maximum signature size."""
    @property
    def block_size(self) -> int:
        """Preferred integral size of messages."""
    @property
    def security_trailer(self) -> int:
        """Size of the security trailer appended to messages."""

def query_context_attributes(
    context: SecurityContext,
    attribute: type[T],
) -> T:
    """Queries an attribute of a security context.

    Enables a transport application to query a security package for certain
    attributes of a security context.

    The attribute must be a type that is a subclass of :class:`SecPkgContext`.
    The instance is created, populated, and returned by this function.

    The following attributes have been implemented:

        :class:`SecPkgContextNames`
        :class:`SecPkgContextPackageInfo`
        :class:`SecPkgContextSessionKey`
        :class:`SecPkgContextSizes`

    This wraps the `QueryContextAttributes`_ Win32 function.

    Args:
        context: The security context to query.
        attribute: The attribute type to query and return.

    Returns:
        The instance of the attribute type provided.

    Raises:
        WindowsError: If the function failed.

    .. _QueryContextAttributes:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/querycontextattributes--general
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from ._win32_types cimport *


cdef class SecurityContext:
    cdef CtxtHandle raw
    cdef TimeStamp raw_expiry
    cdef int _needs_free



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum

from ._security_buffer import SecBufferDesc
from ._security_context import SecurityContext

class QopFlags(enum.IntFlag):
    """Quality of Protection flags for :meth:`encrypt_message`."""

    KERB_WRAP_NO_ENCRYPT = ...
    """
    Produce a header or trailer but do not encrypt the message. Note this is
    the same as ``SECQOP_WRAP_NO_ENCRYPT``.
    """
    SECQOP_WRAP_NO_ENCRYPT = ...
    """
    Produce a header or trailer but do not encrypt the message. Note this is
    the same as ``KERB_WRAP_NO_ENCRYPT``.
    """
    SECQOP_WRAP_OOB_DATA = ...
    """Send an Schannel alert message."""

def decrypt_message(
    context: SecurityContext,
    message: SecBufferDesc,
    seq_no: int,
) -> int:
    """Decrypts a message.

    The function decrypts a message. The message buffers and encryption
    algorithms used is dependent on the security context protocol and the peer
    the output message is being exchanged with.

    The function will decrypt the message buffers in place which means the
    input bytearray/memoryview used to create the :class:`SecBuffer` will also
    be modified. Use the ``data`` property of the relevant buffer to get a copy
    of the decrypted message, otherwise use the ``count`` property to get the
    decrypted bytes length from the input buffer bytearray.

    This wraps the `DecryptMessage`_ Win32 function.

    Args:
        context: The security context used to decrypt the message.
        message: The message to decrypt.
        seq_no: The sequence number that the transport application assigned to
            the message.

    Returns:
        The package specific quality of protection value.

    .. _DecryptMessage:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/decryptmessage--general
    """

def encrypt_message(
    context: SecurityContext,
    qop: QopFlags | int,
    message: SecBufferDesc,
    seq_no: int,
) -> None:
    """Encrypts a message.

    The function encrypts a message to provider privacy. The message buffers
    and encryption algorithms used is dependent on the security context
    protocol and the peer the output message is being exchanged with.

    The function will encrypt the message buffers in place which means the
    input bytearray/memoryview used to create the :class:`SecBuffer` will also
    be modified. Use the ``data`` property of the relevant buffer to get a copy
    of the encrypted message, otherwise use the ``count`` property to get the
    encrypted bytes length from the input buffer bytearray.

    This wraps the `EncryptMessage`_ Win32 function.

    Args:
        context: The security context used to encrypt the message.
        qop: Package specific flags to indicate the quality of protection.
        message: The message to encrypt.
        seq_no: The sequence number that the transport application assigned to
            the message.

    .. _EncryptMessage:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/encryptmessage--general
    """

def make_signature(
    context: SecurityContext,
    qop: QopFlags | int,
    message: SecBufferDesc,
    seq_no: int,
) -> None:
    """Signs a message.

    The function generates a cryptographic checksum of the message, and also
    includes sequencing information to prevent message loss or insertion. The
    message buffers and cryptographic algorithms used are dependent on the
    security context protocol and the peer the output message is being
    exchanged with.

    This wraps the `MakeSignature`_ Win32 function.

    Args:
        context: The security context used to sign the message.
        qop: Package specific flags to indicate the quality of protection.
        message: The message to encrypt.
        seq_no: The sequence number that the transport application assigned to
            the message.

    .. _MakeSignature:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-makesignature
    """

def verify_signature(
    context: SecurityContext,
    message: SecBufferDesc,
    seq_no: int,
) -> int:
    """Verifies a message.

    Verifies that a message signed by :meth:`make_signature` was received in
    the correct sequence and has not been modified. The message buffers and
    cryptographic algorithms used is dependent on the security context protocol
    and the peer the input message was from.

    This wraps the `VerifySignature`_ Win32 function.

    Args:
        context: The security context used to verify the message.
        message: The message to verify.
        seq_no: The sequence number that the transport application assigned to
            the message.

    Returns:
        The package specific quality of protection value.

    .. _VerifySignature:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-verifysignature
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import socket

import pytest

import sspi


def test_sec_exchange_default_cred() -> None:
    spn = f"host/{socket.gethostname()}"

    c_cred = sspi.acquire_credentials_handle(None, "NTLM", sspi.CredentialUse.SECPKG_CRED_OUTBOUND)
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=sspi.IscReq.ISC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=sspi.AscReq.ASC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=sspi.IscReq.ISC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=sspi.AscReq.ASC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=None,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_E_OK


def test_sec_exchange_prealloc_mem() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    c_cred = sspi.acquire_credentials_handle(None, "NTLM", sspi.CredentialUse.SECPKG_CRED_OUTBOUND)
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=None,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_E_OK


def test_invalid_credential() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    invalid_cred = sspi.WinNTAuthIdentity(username="invalid", password="invalid")
    c_cred = sspi.acquire_credentials_handle(
        None,
        "NTLM",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=invalid_cred,
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    with pytest.raises(WindowsError) as e:
        sspi.accept_security_context(
            credential=s_cred,
            context=s_res.context,
            input_buffers=s_input_buffers,
            context_req=0,
            target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
            output_buffers=None,
        )

    assert e.value.winerror == -2146893044  # SEC_E_LOGON_DENIED


def test_sec_exchange_with_channel_binding() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)
    channel_bindings = sspi.SecChannelBindings(
        application_data=b"tls-server-end-point:" + (b"\x00" * 32),
    ).get_sec_buffer_copy()

    c_cred = sspi.acquire_credentials_handle(None, "NTLM", sspi.CredentialUse.SECPKG_CRED_OUTBOUND)
    c_input_buffers = sspi.SecBufferDesc([channel_bindings])
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=None,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_E_OK


def test_sec_exchange_allowed_package_list() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    auth_data = sspi.WinNTAuthIdentity(package_list="NTLM")
    c_cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=None,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_E_OK


def test_sec_exchange_with_negotiate_restricted_package() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    auth_data = sspi.WinNTAuthIdentity(package_list="!Kerberos")
    c_cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_cred = sspi.acquire_credentials_handle(None, "Negotiate", sspi.CredentialUse.SECPKG_CRED_INBOUND)
    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert isinstance(c_res.context.expiry, int)
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sspi.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sspi.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(c_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sspi.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sspi.AcceptContextResult)
    assert isinstance(s_res.context, sspi.AcceptorSecurityContext)
    assert isinstance(s_res.context.context_attr, sspi.AscRet)
    assert s_res.context.expiry > 0
    assert isinstance(s_res.result, sspi.NtStatus)
    assert s_res.result == sspi.NtStatus.SEC_E_OK

    c_input_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(s_output_buffers[0].dangerous_get_view(), sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(buffer, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sspi.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sspi.InitializeContextResult)
    assert isinstance(c_res.context, sspi.InitiatorSecurityContext)
    assert isinstance(c_res.context.context_attr, sspi.IscRet)
    assert c_res.context.expiry > 0
    assert isinstance(c_res.result, sspi.NtStatus)
    assert c_res.result == sspi.NtStatus.SEC_E_OK

    assert c_output_buffers[0].count == 0
    assert c_output_buffers[0].data == b""


def test_fail_with_negotiate_and_no_available_packages() -> None:
    spn = f"host/{socket.gethostname()}"

    auth_data = sspi.WinNTAuthIdentity(package_list="NegoExtender")
    c_cred = sspi.acquire_credentials_handle(
        None,
        "Negotiate",
        sspi.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )
    c_output_buffers = sspi.SecBufferDesc(
        [
            sspi.SecBuffer(None, sspi.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    with pytest.raises(WindowsError) as e:
        sspi.initialize_security_context(
            credential=c_cred,
            context=None,
            target_name=spn,
            context_req=sspi.IscReq.ISC_REQ_ALLOCATE_MEMORY,
            target_data_rep=sspi.TargetDataRep.SECURITY_NATIVE_DREP,
            input_buffers=None,
            output_buffers=c_output_buffers,
        )

    assert e.value.winerror == -2146893042  # SEC_E_NO_CREDENTIALS



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

from cpython.mem cimport PyMem_Free
from cpython.ref cimport PyObject


cdef extern from "Python.h":
    Py_ssize_t PyUnicode_AsWideChar(object unicode, wchar_t *w, Py_ssize_t size)
    wchar_t *PyUnicode_AsWideCharString(object unicode, Py_ssize_t *size)
    PyObject *PyUnicode_FromWideChar(const wchar_t *w, Py_ssize_t size)


cdef class WideCharString:
    # cdef wchar_t *raw
    # cdef Py_ssize_t length

    def __cinit__(WideCharString self, str value) -> None:
        if value is not None:
            self.raw = PyUnicode_AsWideCharString(value, &self.length)
            if self.raw == NULL:
                raise MemoryError()

    def __dealloc__(WideCharString self) -> None:
        if self.raw != NULL:
            PyMem_Free(self.raw)

        self.raw = NULL
        self.length = 0


cdef str wide_char_to_str(
    const wchar_t *value,
    int size = -1,
    int none_is_empty = 0,
):
    if value == NULL:
        return "" if none_is_empty else None

    return <object>PyUnicode_FromWideChar(value, size)



# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from sspi._win32_types cimport *


cdef extern from "Security.h":
    cdef struct _SecBuffer:
        unsigned long cbBuffer
        unsigned long BufferType
        void *pvBuffer
    ctypedef _SecBuffer *PSecBuffer

    cdef struct _SecBufferDesc:
        unsigned long ulVersion
        unsigned long cBuffers
        PSecBuffer pBuffers
    ctypedef _SecBufferDesc *PSecBufferDesc

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-freecontextbuffer
    SECURITY_STATUS FreeContextBuffer(
        PVOID pvContextBuffer
    ) nogil


cdef class SecBufferDesc:
    cdef _SecBufferDesc raw
    cdef list[SecBuffer] _buffers

    cdef void mark_as_allocated(SecBufferDesc self)
    cdef void sync_buffers(SecBufferDesc self)

cdef class SecBuffer:
    cdef _SecBuffer raw
    cdef unsigned char[:] _buffer
    cdef int _needs_free



#!/usr/bin/env python

# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import ctypes
import ctypes.util
import os
import pathlib
import typing

from Cython.Build import cythonize
from setuptools import Extension, setup

SKIP_EXTENSIONS = os.environ.get("SSPI_SKIP_EXTENSIONS", "false").lower() == "true"
SKIP_MODULE_CHECK = os.environ.get("SSPI_SKIP_MODULE_CHECK", "false").lower() == "true"
CYTHON_LINETRACE = os.environ.get("SSPI_CYTHON_TRACING", "false").lower() == "true"


def make_extension(
    name: str,
    module: ctypes.CDLL,
    canary: typing.Optional[str] = None,
    **kwargs: typing.Any,
) -> Extension | None:
    source = pathlib.Path("src") / (name.replace(".", os.sep) + ".pyx")

    if not SKIP_MODULE_CHECK and canary and not hasattr(module, canary):
        print(f"Skipping {source} as it is not supported by the selected SSPI implementation.")
        return None

    if not source.exists():
        raise FileNotFoundError(source)

    print(f"Compiling {source}")
    return Extension(
        name=name,
        sources=[str(source)],
        **kwargs,
    )


raw_extensions = []
if not SKIP_EXTENSIONS:
    sspi_path = pathlib.Path(os.environ.get("SystemRoot", r"C:\Windows")) / "System32" / "Secur32.dll"
    print(f"Using {sspi_path} as SSPI module for platform checks")
    sspi = ctypes.CDLL(str(sspi_path.absolute()))

    for e in [
        "context_attributes",
        "credential_attributes",
        "credential",
        "message",
        "ntstatus",
        "security_buffer",
        "security_context",
        "security_package",
        "text",
    ]:
        name = e
        libraries = ["Secur32"]
        canary = None
        if isinstance(e, tuple):
            name = e[0]
            if len(e) > 1:
                libraries = e[1]
            if len(e) > 2:
                canary = e[2]

        ext = make_extension(
            f"sspi._{name}",
            module=sspi,
            canary=canary,
            libraries=libraries,
            define_macros=[
                ("UNICODE", "1"),
                ("_UNICODE", "1"),
                ("SECURITY_WIN32", "1"),
            ],
        )
        if ext:
            raw_extensions.append(ext)

setup(
    ext_modules=cythonize(
        raw_extensions,
        language_level=3,
        compiler_directives={"linetrace": CYTHON_LINETRACE},
    ),
)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum
import typing as t

SECBUFFER_VERSION: int = 0  #: Constant for SecBufferDesc version

class SecBufferFlags(enum.IntFlag):
    """Security buffer flags."""

    SECBUFFER_NONE = ...
    """No flags are set."""
    SECBUFFER_ATTRMASK = ...
    """Bitmask for the security buffer flags."""
    SECBUFFER_READONLY = ...
    """The buffer is read-only with no checksum."""
    SECBUFFER_READONLY_WITH_CHECKSUM = ...
    """The buffer is read-only with a checksum."""
    SECBUFFER_RESERVED = ...
    """Flags reserved to the security system."""

class SecBufferType(enum.IntEnum):
    """The type of security buffer."""

    SECBUFFER_EMPTY = ...
    """This is a placeholder in the buffer array."""
    SECBUFFER_DATA = ...
    """The buffer contains common data."""
    SECBUFFER_TOKEN = ...
    """The buffer contains the security token portion of the message."""
    SECBUFFER_PKG_PARAMS = ...
    """These are transport-to-package-specific parameters."""
    SECBUFFER_MISSING = ...
    """The security package uses this value to indicate the number of missing bytes in a particular message."""
    SECBUFFER_EXTRA = ...
    """The security package uses this value to indicate the number of extra or unprocessed bytes in a message."""
    SECBUFFER_STREAM_TRAILER = ...
    """The buffer contains a protocol-specific trailer for a particular record."""
    SECBUFFER_STREAM_HEADER = ...
    """The buffer contains a protocol-specific header for a particular record."""
    SECBUFFER_NEGOTIATION_INFO = ...
    """The buffer contains hints from the negotiation package."""
    SECBUFFER_PADDING = ...
    """The buffer contains non-data padding."""
    SECBUFFER_STREAM = ...
    """The buffer contains the whole encrypted message."""
    SECBUFFER_MECHLIST = ...
    """The buffer contains a protocol-specific list of OIDs."""
    SECBUFFER_MECHLIST_SIGNATURE = ...
    """THe buffer contains a signature of the SECBUFFER_MECHLIST buffer."""
    SECBUFFER_TARGET = ...
    """This flag is reserved. Do not use it."""
    SECBUFFER_CHANNEL_BINDINGS = ...
    """The buffer contains channel binding information."""
    SECBUFFER_CHANGE_PASS_RESPONSE = ...
    """The buffer contains a DOMAIN_PASSWORD_INFORMATION structure."""
    SECBUFFER_TARGET_HOST = ...
    """The buffer specifies the SPN of the target."""
    SECBUFFER_ALERT = ...
    """The buffer contains an alert message."""
    SECBUFFER_APPLICATION_PROTOCOLS = ...
    """The buffer contains a list of application protocol IDs."""
    SECBUFFER_SRTP_PROTECTION_PROFILES = ...
    """The buffer contains the list of SRTP protection profiles."""
    SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER = ...
    """The buffer contains the SRTP master key identitifer."""
    SECBUFFER_TOKEN_BINDING = ...
    """The buffer contains the supported token binding protocol version and key parameters."""
    SECBUFFER_PRESHARED_KEY = ...
    """The buffer contains the preshared key."""
    SECBUFFER_PRESHARED_KEY_IDENTITY = ...
    """The buffer contains the preshared key identity."""
    SECBUFFER_DTLS_MTU = ...
    """The buffer contains the setting for the maximum transmission unit size of DTLS."""
    SECBUFFER_SEND_GENERIC_TLS_EXTENSION = ...
    """The buffer contains generic TLS extensions for sending."""
    SECBUFFER_SUBSCRIBE_GENERIC_TLS_EXTENSION = ...
    """The buffer contains generic TLS extensions for subscribing."""
    SECBUFFER_FLAGS = ...
    """The buffer contains the ISC/ASC REQ flags."""
    SECBUFFER_TRAFFIC_SECRETS = ...
    """The buffer contains message sequence lengths and corresponding traffic secrets."""
    SECBUFFER_CERTIFICATE_REQUEST_CONTEXT = ...
    """The buffer contains the TLS 1.3 certificate request context."""

class SecBufferDesc(t.Iterable):
    """An array of SecBuffer structures.

    This type is used to encapsulate an array of :class:`SecBuffer` instances
    used for exchanging data with a security package. The array size and values
    are immutable once created.

    Args:
        buffers: A list of SecBuffer values to include in this array.
        version: The version number of the structure, defaults to
            SECBUFFER_VERSION.
    """

    def __init__(
        self,
        buffers: list[SecBuffer],
        *,
        version: int = SECBUFFER_VERSION,
    ) -> None: ...
    def __iter__(self) -> t.Iterator[SecBuffer]:
        """Creates an iterable of the contained buffers."""
    def __len__(self) -> int:
        """Returns the number of buffers in this structure."""
    def __getitem__(self, key: int) -> SecBuffer:
        """Gets the buffer at the specified index."""
    @property
    def version(self) -> int:
        """The version number of the structure."""

class SecBuffer:
    """A security package buffer.

    A buffer used to exchange data to and from a security package operation.
    Depending on the buffer type and operation, the data might be mutated in
    place by the security package.

    The ``data`` property will create a copy of the buffer bytes while
    :meth:`dangerous_get_view` can be used to get a view of the data pointed by
    the buffer without any copying. Be careful not to use the view returned
    once the buffer has been freed.

    Args:
        data: A bytearray or memoryview to a writable bytes buffer containing
            the buffer data. Can be None if using an empty data buffer.
        buffer_type: The buffer type.
        buffer_flags: Any buffer flags to associate with the data, defaults to
            no flags.
    """

    def __init__(
        self,
        data: bytearray | memoryview | None,
        buffer_type: SecBufferType | int,
        buffer_flags: SecBufferFlags | int = 0,
    ) -> None: ...
    @property
    def count(self) -> int:
        """The length of the buffer."""
    @property
    def data(self) -> bytes:
        """A copy of the buffer bytes."""
    @property
    def buffer_type(self) -> SecBufferType:
        """The buffer type."""
    @property
    def buffer_flags(self) -> SecBufferFlags:
        """The buffer flags."""
    def dangerous_get_view(self) -> memoryview:
        """The buffer memoryview.

        This returns a memoryview to the raw buffer bytes in memory. While this
        does not copy the data from the buffer, it is only valid if the class
        instance has not been deallocated. As the memoryview is a reference to
        unmanaged memory it is critical that the view is not used once the
        buffer is deallocated. To get a copy of the bytes in a safer fashion
        use :meth:`data`.
        """

class SecChannelBindings:
    """Channel binding information.

    Specifies channel binding information for a security context. SSPI only
    uses the application_data property but others might used for interop with
    other implementations. Any value passed in will be copied to an internal
    buffer as SSPI requires it to be placed in contiguous memory. The properties
    will return a copy of the values if set as well.

    The methods :meth:`get_sec_buffer_copy` and :meth:`dangerous_get_buffer`
    can be used to get a :class:`SecBuffer` that contains this structure value.
    The former copies the underlying bytes so is less efficient but is safer to
    use. The latter method uses a reference so the returned buffer will only be
    valid until the instance it came from was not deallocated.

    This wraps the `SEC_CHANNEL_BINDINGS`_ Win32 structure.

    .. _SEC_CHANNEL_BINDINGS:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-sec_channel_bindings

    Args:
        initiator_addr_type: The type of the initiator/client address.
        initiator_addr: The initiator address.
        acceptor_addr_type: The type of the acceptor/server address.
        acceptor_addr: The acceptor address.
        application_data: The channel binding data.
    """

    def __init__(
        self,
        *,
        initiator_addr_type: int = 0,
        initiator_addr: bytes | bytearray | memoryview | None = None,
        acceptor_addr_type: int = 0,
        acceptor_addr: bytes | bytearray | memoryview | None = None,
        application_data: bytes | bytearray | memoryview | None = None,
    ) -> None: ...
    @property
    def initiator_addr_type(self) -> int:
        """The initiator address type."""
    @property
    def initiator_addr(self) -> bytes | None:
        """A copy of the initiator address or None if not present."""
    @property
    def acceptor_addr_type(self) -> int:
        """The acceptor address type."""
    @property
    def acceptor_addr(self) -> bytes | None:
        """A copy of the acceptor address or None if not present."""
    @property
    def application_data(self) -> bytes | None:
        """A copy of the application data or None if not present."""
    def get_sec_buffer_copy(self) -> SecBuffer:
        """The SecBuffer value.

        This returns a :class:`SecBuffer` object that can be used as part of a
        security input buffer. Unlike :meth:`dangerous_get_sec_buffer` it will
        create the buffer with a copy of the underlying bytes making it safe to
        use beyond the lifetime of the instance it came from.

        Use :meth:`dangerous_get_sec_buffer` to use a more efficient method of
        creating the buffer (no memory copies) if you can guarantee it will
        only be used when this instance is still referenced and not
        deallocated.

        Returns:
            SecBuffer: The security buffer with the copied memory.
        """
    def dangerous_get_sec_buffer(self) -> SecBuffer:
        """The SecBuffer value.

        This returns a :class:`SecBuffer` object that can be used as part of a
        security input buffer. It is marked as dangerous as it does not copy
        the underlying bytes but uses a pointer to the unmanaged memory for the
        buffer. It is critical the returned buffer is not used once this object
        has been deallocated as it will invalidate the memory the buffer points
        to.

        A safer alternative is to use the :meth:`get_sec_buffer_copy` to get a
        :class:`SecBuffer` that contains the copied struct bytes rather than a
        reference.

        Returns:
            SecBuffer: The security buffer with the referenced memory.
        """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import collections
import enum

from cpython.exc cimport PyErr_SetFromWindowsErr

from sspi._credential cimport Credential
from sspi._security_buffer cimport PSecBufferDesc, SecBufferDesc
from sspi._text cimport WideCharString
from sspi._win32_types cimport *

from sspi._ntstatus import NtStatus


cdef extern from "Security.h":
    unsigned long _SECURITY_NATIVE_DREP "SECURITY_NATIVE_DREP"
    unsigned long _SECURITY_NETWORK_DREP "SECURITY_NETWORK_DREP"

    unsigned long _ASC_REQ_DELEGATE "ASC_REQ_DELEGATE"
    unsigned long _ASC_REQ_MUTUAL_AUTH "ASC_REQ_MUTUAL_AUTH"
    unsigned long _ASC_REQ_REPLAY_DETECT "ASC_REQ_REPLAY_DETECT"
    unsigned long _ASC_REQ_SEQUENCE_DETECT "ASC_REQ_SEQUENCE_DETECT"
    unsigned long _ASC_REQ_CONFIDENTIALITY "ASC_REQ_CONFIDENTIALITY"
    unsigned long _ASC_REQ_USE_SESSION_KEY "ASC_REQ_USE_SESSION_KEY"
    unsigned long _ASC_REQ_SESSION_TICKET "ASC_REQ_SESSION_TICKET"
    unsigned long _ASC_REQ_ALLOCATE_MEMORY "ASC_REQ_ALLOCATE_MEMORY"
    unsigned long _ASC_REQ_USE_DCE_STYLE "ASC_REQ_USE_DCE_STYLE"
    unsigned long _ASC_REQ_DATAGRAM "ASC_REQ_DATAGRAM"
    unsigned long _ASC_REQ_CONNECTION "ASC_REQ_CONNECTION"
    unsigned long _ASC_REQ_CALL_LEVEL "ASC_REQ_CALL_LEVEL"
    unsigned long _ASC_REQ_FRAGMENT_SUPPLIED "ASC_REQ_FRAGMENT_SUPPLIED"
    unsigned long _ASC_REQ_EXTENDED_ERROR "ASC_REQ_EXTENDED_ERROR"
    unsigned long _ASC_REQ_STREAM "ASC_REQ_STREAM"
    unsigned long _ASC_REQ_INTEGRITY "ASC_REQ_INTEGRITY"
    unsigned long _ASC_REQ_LICENSING "ASC_REQ_LICENSING"
    unsigned long _ASC_REQ_IDENTIFY "ASC_REQ_IDENTIFY"
    unsigned long _ASC_REQ_ALLOW_NULL_SESSION "ASC_REQ_ALLOW_NULL_SESSION"
    unsigned long _ASC_REQ_ALLOW_NON_USER_LOGONS "ASC_REQ_ALLOW_NON_USER_LOGONS"
    unsigned long _ASC_REQ_ALLOW_CONTEXT_REPLAY "ASC_REQ_ALLOW_CONTEXT_REPLAY"
    unsigned long _ASC_REQ_FRAGMENT_TO_FIT "ASC_REQ_FRAGMENT_TO_FIT"
    unsigned long _ASC_REQ_NO_TOKEN "ASC_REQ_NO_TOKEN"
    unsigned long _ASC_REQ_PROXY_BINDINGS "ASC_REQ_PROXY_BINDINGS"
    unsigned long _ASC_REQ_ALLOW_MISSING_BINDINGS "ASC_REQ_ALLOW_MISSING_BINDINGS"

    unsigned long _ASC_RET_DELEGATE "ASC_RET_DELEGATE"
    unsigned long _ASC_RET_MUTUAL_AUTH "ASC_RET_MUTUAL_AUTH"
    unsigned long _ASC_RET_REPLAY_DETECT "ASC_RET_REPLAY_DETECT"
    unsigned long _ASC_RET_SEQUENCE_DETECT "ASC_RET_SEQUENCE_DETECT"
    unsigned long _ASC_RET_CONFIDENTIALITY "ASC_RET_CONFIDENTIALITY"
    unsigned long _ASC_RET_USE_SESSION_KEY "ASC_RET_USE_SESSION_KEY"
    unsigned long _ASC_RET_SESSION_TICKET "ASC_RET_SESSION_TICKET"
    unsigned long _ASC_RET_ALLOCATED_MEMORY "ASC_RET_ALLOCATED_MEMORY"
    unsigned long _ASC_RET_USED_DCE_STYLE "ASC_RET_USED_DCE_STYLE"
    unsigned long _ASC_RET_DATAGRAM "ASC_RET_DATAGRAM"
    unsigned long _ASC_RET_CONNECTION "ASC_RET_CONNECTION"
    unsigned long _ASC_RET_CALL_LEVEL "ASC_RET_CALL_LEVEL"
    unsigned long _ASC_RET_THIRD_LEG_FAILED "ASC_RET_THIRD_LEG_FAILED"
    unsigned long _ASC_RET_EXTENDED_ERROR "ASC_RET_EXTENDED_ERROR"
    unsigned long _ASC_RET_STREAM "ASC_RET_STREAM"
    unsigned long _ASC_RET_INTEGRITY "ASC_RET_INTEGRITY"
    unsigned long _ASC_RET_LICENSING "ASC_RET_LICENSING"
    unsigned long _ASC_RET_IDENTIFY "ASC_RET_IDENTIFY"
    unsigned long _ASC_RET_NULL_SESSION "ASC_RET_NULL_SESSION"
    unsigned long _ASC_RET_ALLOW_NON_USER_LOGONS "ASC_RET_ALLOW_NON_USER_LOGONS"
    unsigned long _ASC_RET_ALLOW_CONTEXT_REPLAY "ASC_RET_ALLOW_CONTEXT_REPLAY"
    unsigned long _ASC_RET_FRAGMENT_ONLY "ASC_RET_FRAGMENT_ONLY"
    unsigned long _ASC_RET_NO_TOKEN "ASC_RET_NO_TOKEN"
    unsigned long _ASC_RET_NO_ADDITIONAL_TOKEN "ASC_RET_NO_ADDITIONAL_TOKEN"

    unsigned long _ISC_REQ_DELEGATE "ISC_REQ_DELEGATE"
    unsigned long _ISC_REQ_MUTUAL_AUTH "ISC_REQ_MUTUAL_AUTH"
    unsigned long _ISC_REQ_REPLAY_DETECT "ISC_REQ_REPLAY_DETECT"
    unsigned long _ISC_REQ_SEQUENCE_DETECT "ISC_REQ_SEQUENCE_DETECT"
    unsigned long _ISC_REQ_CONFIDENTIALITY "ISC_REQ_CONFIDENTIALITY"
    unsigned long _ISC_REQ_USE_SESSION_KEY "ISC_REQ_USE_SESSION_KEY"
    unsigned long _ISC_REQ_PROMPT_FOR_CREDS "ISC_REQ_PROMPT_FOR_CREDS"
    unsigned long _ISC_REQ_USE_SUPPLIED_CREDS "ISC_REQ_USE_SUPPLIED_CREDS"
    unsigned long _ISC_REQ_ALLOCATE_MEMORY "ISC_REQ_ALLOCATE_MEMORY"
    unsigned long _ISC_REQ_USE_DCE_STYLE "ISC_REQ_USE_DCE_STYLE"
    unsigned long _ISC_REQ_DATAGRAM "ISC_REQ_DATAGRAM"
    unsigned long _ISC_REQ_CONNECTION "ISC_REQ_CONNECTION"
    unsigned long _ISC_REQ_CALL_LEVEL "ISC_REQ_CALL_LEVEL"
    unsigned long _ISC_REQ_FRAGMENT_SUPPLIED "ISC_REQ_FRAGMENT_SUPPLIED"
    unsigned long _ISC_REQ_EXTENDED_ERROR "ISC_REQ_EXTENDED_ERROR"
    unsigned long _ISC_REQ_STREAM "ISC_REQ_STREAM"
    unsigned long _ISC_REQ_INTEGRITY "ISC_REQ_INTEGRITY"
    unsigned long _ISC_REQ_IDENTIFY "ISC_REQ_IDENTIFY"
    unsigned long _ISC_REQ_NULL_SESSION "ISC_REQ_NULL_SESSION"
    unsigned long _ISC_REQ_MANUAL_CRED_VALIDATION "ISC_REQ_MANUAL_CRED_VALIDATION"
    unsigned long _ISC_REQ_RESERVED1 "ISC_REQ_RESERVED1"
    unsigned long _ISC_REQ_FRAGMENT_TO_FIT "ISC_REQ_FRAGMENT_TO_FIT"
    unsigned long _ISC_REQ_FORWARD_CREDENTIALS "ISC_REQ_FORWARD_CREDENTIALS"
    unsigned long _ISC_REQ_NO_INTEGRITY "ISC_REQ_NO_INTEGRITY"
    unsigned long _ISC_REQ_USE_HTTP_STYLE "ISC_REQ_USE_HTTP_STYLE"
    unsigned long _ISC_REQ_UNVERIFIED_TARGET_NAME "ISC_REQ_UNVERIFIED_TARGET_NAME"
    unsigned long _ISC_REQ_CONFIDENTIALITY_ONLY "ISC_REQ_CONFIDENTIALITY_ONLY"

    unsigned long _ISC_RET_DELEGATE "ISC_RET_DELEGATE"
    unsigned long _ISC_RET_MUTUAL_AUTH "ISC_RET_MUTUAL_AUTH"
    unsigned long _ISC_RET_REPLAY_DETECT "ISC_RET_REPLAY_DETECT"
    unsigned long _ISC_RET_SEQUENCE_DETECT "ISC_RET_SEQUENCE_DETECT"
    unsigned long _ISC_RET_CONFIDENTIALITY "ISC_RET_CONFIDENTIALITY"
    unsigned long _ISC_RET_USE_SESSION_KEY "ISC_RET_USE_SESSION_KEY"
    unsigned long _ISC_RET_USED_COLLECTED_CREDS "ISC_RET_USED_COLLECTED_CREDS"
    unsigned long _ISC_RET_USED_SUPPLIED_CREDS "ISC_RET_USED_SUPPLIED_CREDS"
    unsigned long _ISC_RET_ALLOCATED_MEMORY "ISC_RET_ALLOCATED_MEMORY"
    unsigned long _ISC_RET_USED_DCE_STYLE "ISC_RET_USED_DCE_STYLE"
    unsigned long _ISC_RET_DATAGRAM "ISC_RET_DATAGRAM"
    unsigned long _ISC_RET_CONNECTION "ISC_RET_CONNECTION"
    unsigned long _ISC_RET_INTERMEDIATE_RETURN "ISC_RET_INTERMEDIATE_RETURN"
    unsigned long _ISC_RET_CALL_LEVEL "ISC_RET_CALL_LEVEL"
    unsigned long _ISC_RET_EXTENDED_ERROR "ISC_RET_EXTENDED_ERROR"
    unsigned long _ISC_RET_STREAM "ISC_RET_STREAM"
    unsigned long _ISC_RET_INTEGRITY "ISC_RET_INTEGRITY"
    unsigned long _ISC_RET_IDENTIFY "ISC_RET_IDENTIFY"
    unsigned long _ISC_RET_NULL_SESSION "ISC_RET_NULL_SESSION"
    unsigned long _ISC_RET_MANUAL_CRED_VALIDATION "ISC_RET_MANUAL_CRED_VALIDATION"
    unsigned long _ISC_RET_RESERVED1 "ISC_RET_RESERVED1"
    unsigned long _ISC_RET_FRAGMENT_ONLY "ISC_RET_FRAGMENT_ONLY"
    unsigned long _ISC_RET_FORWARD_CREDENTIALS "ISC_RET_FORWARD_CREDENTIALS"
    unsigned long _ISC_RET_USED_HTTP_STYLE "ISC_RET_USED_HTTP_STYLE"
    unsigned long _ISC_RET_NO_ADDITIONAL_TOKEN "ISC_RET_NO_ADDITIONAL_TOKEN"
    unsigned long _ISC_RET_REAUTHENTICATION "ISC_RET_REAUTHENTICATION"
    unsigned long _ISC_RET_CONFIDENTIALITY_ONLY "ISC_RET_CONFIDENTIALITY_ONLY"

    # https://learn.microsoft.com/en-us/windows/win32/secauthn/acceptsecuritycontext--general
    SECURITY_STATUS AcceptSecurityContext(
        PCredHandle    phCredential,
        PCtxtHandle    phContext,
        PSecBufferDesc pInput,
        unsigned long  fContextReq,
        unsigned long  TargetDataRep,
        PCtxtHandle    phNewContext,
        PSecBufferDesc pOutput,
        unsigned long  *pfContextAttr,
        PTimeStamp     ptsExpiry
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-completeauthtoken
    SECURITY_STATUS CompleteAuthToken(
        PCtxtHandle    phContext,
        PSecBufferDesc pToken
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-initializesecuritycontextw
    SECURITY_STATUS InitializeSecurityContextW(
        PCredHandle    phCredential,
        PCtxtHandle    phContext,
        LPWSTR         pTargetName,
        unsigned long  fContextReq,
        unsigned long  Reserved1,
        unsigned long  TargetDataRep,
        PSecBufferDesc pInput,
        unsigned long  Reserved2,
        PCtxtHandle    phNewContext,
        PSecBufferDesc pOutput,
        unsigned long  *pfContextAttr,
        PTimeStamp     ptsExpiry
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-deletesecuritycontext
    SECURITY_STATUS DeleteSecurityContext(
        PCtxtHandle phContext
    ) nogil

class TargetDataRep(enum.IntEnum):
    SECURITY_NATIVE_DREP = _SECURITY_NATIVE_DREP
    SECURITY_NETWORK_DREP = _SECURITY_NETWORK_DREP

class AscReq(enum.IntFlag):
    ASC_REQ_DELEGATE = _ASC_REQ_DELEGATE
    ASC_REQ_MUTUAL_AUTH = _ASC_REQ_MUTUAL_AUTH
    ASC_REQ_REPLAY_DETECT = _ASC_REQ_REPLAY_DETECT
    ASC_REQ_SEQUENCE_DETECT = _ASC_REQ_SEQUENCE_DETECT
    ASC_REQ_CONFIDENTIALITY = _ASC_REQ_CONFIDENTIALITY
    ASC_REQ_USE_SESSION_KEY = _ASC_REQ_USE_SESSION_KEY
    ASC_REQ_SESSION_TICKET = _ASC_REQ_SESSION_TICKET
    ASC_REQ_ALLOCATE_MEMORY = _ASC_REQ_ALLOCATE_MEMORY
    ASC_REQ_USE_DCE_STYLE = _ASC_REQ_USE_DCE_STYLE
    ASC_REQ_DATAGRAM = _ASC_REQ_DATAGRAM
    ASC_REQ_CONNECTION = _ASC_REQ_CONNECTION
    ASC_REQ_CALL_LEVEL = _ASC_REQ_CALL_LEVEL
    ASC_REQ_FRAGMENT_SUPPLIED = _ASC_REQ_FRAGMENT_SUPPLIED
    ASC_REQ_EXTENDED_ERROR = _ASC_REQ_EXTENDED_ERROR
    ASC_REQ_STREAM = _ASC_REQ_STREAM
    ASC_REQ_INTEGRITY = _ASC_REQ_INTEGRITY
    ASC_REQ_LICENSING = _ASC_REQ_LICENSING
    ASC_REQ_IDENTIFY = _ASC_REQ_IDENTIFY
    ASC_REQ_ALLOW_NULL_SESSION = _ASC_REQ_ALLOW_NULL_SESSION
    ASC_REQ_ALLOW_NON_USER_LOGONS = _ASC_REQ_ALLOW_NON_USER_LOGONS
    ASC_REQ_ALLOW_CONTEXT_REPLAY = _ASC_REQ_ALLOW_CONTEXT_REPLAY
    ASC_REQ_FRAGMENT_TO_FIT = _ASC_REQ_FRAGMENT_TO_FIT
    ASC_REQ_NO_TOKEN = _ASC_REQ_NO_TOKEN
    ASC_REQ_PROXY_BINDINGS = _ASC_REQ_PROXY_BINDINGS
    ASC_REQ_ALLOW_MISSING_BINDINGS = _ASC_REQ_ALLOW_MISSING_BINDINGS

class AscRet(enum.IntFlag):
    ASC_RET_DELEGATE = _ASC_RET_DELEGATE
    ASC_RET_MUTUAL_AUTH = _ASC_RET_MUTUAL_AUTH
    ASC_RET_REPLAY_DETECT = _ASC_RET_REPLAY_DETECT
    ASC_RET_SEQUENCE_DETECT = _ASC_RET_SEQUENCE_DETECT
    ASC_RET_CONFIDENTIALITY = _ASC_RET_CONFIDENTIALITY
    ASC_RET_USE_SESSION_KEY = _ASC_RET_USE_SESSION_KEY
    ASC_RET_SESSION_TICKET = _ASC_RET_SESSION_TICKET
    ASC_RET_ALLOCATED_MEMORY = _ASC_RET_ALLOCATED_MEMORY
    ASC_RET_USED_DCE_STYLE = _ASC_RET_USED_DCE_STYLE
    ASC_RET_DATAGRAM = _ASC_RET_DATAGRAM
    ASC_RET_CONNECTION = _ASC_RET_CONNECTION
    ASC_RET_CALL_LEVEL = _ASC_RET_CALL_LEVEL
    ASC_RET_THIRD_LEG_FAILED = _ASC_RET_THIRD_LEG_FAILED
    ASC_RET_EXTENDED_ERROR = _ASC_RET_EXTENDED_ERROR
    ASC_RET_STREAM = _ASC_RET_STREAM
    ASC_RET_INTEGRITY = _ASC_RET_INTEGRITY
    ASC_RET_LICENSING = _ASC_RET_LICENSING
    ASC_RET_IDENTIFY = _ASC_RET_IDENTIFY
    ASC_RET_NULL_SESSION = _ASC_RET_NULL_SESSION
    ASC_RET_ALLOW_NON_USER_LOGONS = _ASC_RET_ALLOW_NON_USER_LOGONS
    ASC_RET_ALLOW_CONTEXT_REPLAY = _ASC_RET_ALLOW_CONTEXT_REPLAY
    ASC_RET_FRAGMENT_ONLY = _ASC_RET_FRAGMENT_ONLY
    ASC_RET_NO_TOKEN = _ASC_RET_NO_TOKEN
    ASC_RET_NO_ADDITIONAL_TOKEN = _ASC_RET_NO_ADDITIONAL_TOKEN

class IscReq(enum.IntFlag):
    ISC_REQ_DELEGATE = _ISC_REQ_DELEGATE
    ISC_REQ_MUTUAL_AUTH = _ISC_REQ_MUTUAL_AUTH
    ISC_REQ_REPLAY_DETECT = _ISC_REQ_REPLAY_DETECT
    ISC_REQ_SEQUENCE_DETECT = _ISC_REQ_SEQUENCE_DETECT
    ISC_REQ_CONFIDENTIALITY = _ISC_REQ_CONFIDENTIALITY
    ISC_REQ_USE_SESSION_KEY = _ISC_REQ_USE_SESSION_KEY
    ISC_REQ_PROMPT_FOR_CREDS = _ISC_REQ_PROMPT_FOR_CREDS
    ISC_REQ_USE_SUPPLIED_CREDS = _ISC_REQ_USE_SUPPLIED_CREDS
    ISC_REQ_ALLOCATE_MEMORY = _ISC_REQ_ALLOCATE_MEMORY
    ISC_REQ_USE_DCE_STYLE = _ISC_REQ_USE_DCE_STYLE
    ISC_REQ_DATAGRAM = _ISC_REQ_DATAGRAM
    ISC_REQ_CONNECTION = _ISC_REQ_CONNECTION
    ISC_REQ_CALL_LEVEL = _ISC_REQ_CALL_LEVEL
    ISC_REQ_FRAGMENT_SUPPLIED = _ISC_REQ_FRAGMENT_SUPPLIED
    ISC_REQ_EXTENDED_ERROR = _ISC_REQ_EXTENDED_ERROR
    ISC_REQ_STREAM = _ISC_REQ_STREAM
    ISC_REQ_INTEGRITY = _ISC_REQ_INTEGRITY
    ISC_REQ_IDENTIFY = _ISC_REQ_IDENTIFY
    ISC_REQ_NULL_SESSION = _ISC_REQ_NULL_SESSION
    ISC_REQ_MANUAL_CRED_VALIDATION = _ISC_REQ_MANUAL_CRED_VALIDATION
    ISC_REQ_RESERVED1 = _ISC_REQ_RESERVED1
    ISC_REQ_FRAGMENT_TO_FIT = _ISC_REQ_FRAGMENT_TO_FIT
    ISC_REQ_FORWARD_CREDENTIALS = _ISC_REQ_FORWARD_CREDENTIALS
    ISC_REQ_NO_INTEGRITY = _ISC_REQ_NO_INTEGRITY
    ISC_REQ_USE_HTTP_STYLE = _ISC_REQ_USE_HTTP_STYLE
    ISC_REQ_UNVERIFIED_TARGET_NAME = _ISC_REQ_UNVERIFIED_TARGET_NAME
    ISC_REQ_CONFIDENTIALITY_ONLY = _ISC_REQ_CONFIDENTIALITY_ONLY

class IscRet(enum.IntFlag):
    ISC_RET_DELEGATE = _ISC_RET_DELEGATE
    ISC_RET_MUTUAL_AUTH = _ISC_RET_MUTUAL_AUTH
    ISC_RET_REPLAY_DETECT = _ISC_RET_REPLAY_DETECT
    ISC_RET_SEQUENCE_DETECT = _ISC_RET_SEQUENCE_DETECT
    ISC_RET_CONFIDENTIALITY = _ISC_RET_CONFIDENTIALITY
    ISC_RET_USE_SESSION_KEY = _ISC_RET_USE_SESSION_KEY
    ISC_RET_USED_COLLECTED_CREDS = _ISC_RET_USED_COLLECTED_CREDS
    ISC_RET_USED_SUPPLIED_CREDS = _ISC_RET_USED_SUPPLIED_CREDS
    ISC_RET_ALLOCATED_MEMORY = _ISC_RET_ALLOCATED_MEMORY
    ISC_RET_USED_DCE_STYLE = _ISC_RET_USED_DCE_STYLE
    ISC_RET_DATAGRAM = _ISC_RET_DATAGRAM
    ISC_RET_CONNECTION = _ISC_RET_CONNECTION
    ISC_RET_INTERMEDIATE_RETURN = _ISC_RET_INTERMEDIATE_RETURN
    ISC_RET_CALL_LEVEL = _ISC_RET_CALL_LEVEL
    ISC_RET_EXTENDED_ERROR = _ISC_RET_EXTENDED_ERROR
    ISC_RET_STREAM = _ISC_RET_STREAM
    ISC_RET_INTEGRITY = _ISC_RET_INTEGRITY
    ISC_RET_IDENTIFY = _ISC_RET_IDENTIFY
    ISC_RET_NULL_SESSION = _ISC_RET_NULL_SESSION
    ISC_RET_MANUAL_CRED_VALIDATION = _ISC_RET_MANUAL_CRED_VALIDATION
    ISC_RET_RESERVED1 = _ISC_RET_RESERVED1
    ISC_RET_FRAGMENT_ONLY = _ISC_RET_FRAGMENT_ONLY
    ISC_RET_FORWARD_CREDENTIALS = _ISC_RET_FORWARD_CREDENTIALS
    ISC_RET_USED_HTTP_STYLE = _ISC_RET_USED_HTTP_STYLE
    ISC_RET_NO_ADDITIONAL_TOKEN = _ISC_RET_NO_ADDITIONAL_TOKEN
    ISC_RET_REAUTHENTICATION = _ISC_RET_REAUTHENTICATION
    ISC_RET_CONFIDENTIALITY_ONLY = _ISC_RET_CONFIDENTIALITY_ONLY

cdef class SecurityContext:
    # cdef CtxtHandle raw
    # cdef TimeStamp raw_expiry
    # cdef int _needs_free

    def __dealloc__(SecurityContext self):
        if self._needs_free:
            DeleteSecurityContext(&self.raw)
            self._needs_free = 0

    @property
    def expiry(SecurityContext self) -> int:
        return (<unsigned long long>self.raw_expiry.HighPart << 32) | self.raw_expiry.LowPart

cdef class AcceptorSecurityContext(SecurityContext):
    cdef unsigned long raw_context_attr

    @property
    def context_attr(AcceptorSecurityContext self) -> AscRet:
        return AscRet(self.raw_context_attr)

cdef class InitiatorSecurityContext(SecurityContext):
    cdef unsigned long raw_context_attr

    @property
    def context_attr(InitiatorSecurityContext self) -> IscRet:
        return IscRet(self.raw_context_attr)

AcceptContextResult = collections.namedtuple(
    'AcceptContextResult',
    ['context', 'result'],
)

InitializeContextResult = collections.namedtuple(
    'InitializeContextResult',
    ['context', 'result'],
)

def accept_security_context(
    Credential credential,
    AcceptorSecurityContext context,
    SecBufferDesc input_buffers,
    unsigned long context_req,
    unsigned long target_data_rep,
    SecBufferDesc output_buffers,
) -> AcceptContextResult:
    cdef PCredHandle cred_handle = NULL
    if credential:
        cred_handle = &credential.raw

    cdef PCtxtHandle in_context = NULL
    cdef AcceptorSecurityContext out_context = None
    if context:
        in_context = &context.raw
        out_context = context
    else:
        in_context = NULL
        out_context = AcceptorSecurityContext()

    cdef PSecBufferDesc input_buffers_raw = NULL
    if input_buffers:
        input_buffers_raw = &input_buffers.raw
    cdef PSecBufferDesc output_buffers_raw = NULL
    if output_buffers:
        output_buffers_raw = &output_buffers.raw

    with nogil:
        res = AcceptSecurityContext(
            cred_handle,
            in_context,
            input_buffers_raw,
            context_req,
            target_data_rep,
            &out_context.raw,
            output_buffers_raw,
            &out_context.raw_context_attr,
            &out_context.raw_expiry,
        )

    if res not in [
        NtStatus.SEC_I_COMPLETE_AND_CONTINUE,
        NtStatus.SEC_I_COMPLETE_NEEDED,
        NtStatus.SEC_I_CONTINUE_NEEDED,
        NtStatus.SEC_E_OK,
    ]:
        PyErr_SetFromWindowsErr(res)

    if output_buffers:
        output_buffers.sync_buffers()
        if context_req & _ASC_REQ_ALLOCATE_MEMORY:
            output_buffers.mark_as_allocated()
    out_context._needs_free = 1

    return AcceptContextResult(
        context=out_context,
        result=NtStatus(res),
    )

def complete_auth_token(
    SecurityContext context not None,
    SecBufferDesc token not None,
) -> None:
    with nogil:
        res = CompleteAuthToken(
            &context.raw,
            &token.raw,
        )

    if res:
        PyErr_SetFromWindowsErr(res)

def initialize_security_context(
    Credential credential,
    InitiatorSecurityContext context,
    str target_name not None,
    unsigned long context_req,
    unsigned long target_data_rep,
    SecBufferDesc input_buffers,
    SecBufferDesc output_buffers,
) -> InitializeContextResult:
    cdef PCredHandle cred_handle = NULL
    if credential:
        cred_handle = &credential.raw

    cdef PCtxtHandle in_context = NULL
    cdef InitiatorSecurityContext out_context = None
    if context:
        in_context = &context.raw
        out_context = context
    else:
        in_context = NULL
        out_context = InitiatorSecurityContext()

    cdef WideCharString target_name_wchar = WideCharString(target_name)

    cdef PSecBufferDesc input_buffers_raw = NULL
    if input_buffers:
        input_buffers_raw = &input_buffers.raw
    cdef PSecBufferDesc output_buffers_raw = NULL
    if output_buffers:
        output_buffers_raw = &output_buffers.raw

    with nogil:
        res = InitializeSecurityContextW(
            cred_handle,
            in_context,
            target_name_wchar.raw,
            context_req,
            0,
            target_data_rep,
            input_buffers_raw,
            0,
            &out_context.raw,
            output_buffers_raw,
            &out_context.raw_context_attr,
            &out_context.raw_expiry,
        )

    if res not in [
        NtStatus.SEC_I_COMPLETE_AND_CONTINUE,
        NtStatus.SEC_I_COMPLETE_NEEDED,
        NtStatus.SEC_I_CONTINUE_NEEDED,
        NtStatus.SEC_E_OK,
    ]:
        PyErr_SetFromWindowsErr(res)

    if output_buffers:
        output_buffers.sync_buffers()
        if context_req & _ISC_REQ_ALLOCATE_MEMORY:
            output_buffers.mark_as_allocated()
    out_context._needs_free = 1

    return InitializeContextResult(
        context=out_context,
        result=NtStatus(res),
    )
